/*
 * omf_rec: primitive operations to build omf records, and to
 *          write a built record to a file.
 * This module is relatively stateless, in that it doesn't remember
 * things about the object file being built.
 * Although simpler, this makes it harder to support the varied options
 * available in the object format.
 */
 
#include <stdio.h>
#include <stdlib.h>
#include "omf.h"

#ifndef OMF_MINGROW
#define OMF_MINGROW 32
#endif

orec_t *
omf_allocrec(void)
{
	orec_t *r;
	if (!(r = malloc(sizeof *r))) {
		fprintf(stderr,"out of memory in omf_allocrec\n");
		exit(1);
	}
	r->len = r->curp = 0; r->bufp = 0;
	return r;
}

int
omf_freerec(orec_t *r)
{
	if (r->bufp)	free(r->bufp);
	free(r);
	return 0;
}

int
omf_growrec(orec_t *r, int n)
{
	if (r->len <= r->curp+n) {
		uchar_t *p;
		int	l;
		l = max(OMF_MINGROW, n);
		if (!(p = realloc(r->bufp, r->len+l))) {
			return -1;
		}
		r->bufp = p;
		r->len += l;
	}
	return 0;
}

int
omf_putbyte(orec_t *r, int val)
{
	if (omf_growrec(r,1) == -1) {
		fprintf(stderr,"error: no memory in omf_putbyte\n");
		exit(1);
	}
	return r->bufp[r->curp++] = val;
}

int
omf_putword(orec_t *r, int u)
{
	omf_putbyte(r,u);
	omf_putbyte(r,u>>8);
	return 0;
}

int
omf_putdword(orec_t *r, long u)
{
	omf_putword(r,u);
	omf_putword(r,u>>16);
	return 0;
}

int
omf_puteither(orec_t *r,long u)
{
	/* watcom switch here to 16 if !IsMS386 && !IsPharLap */
	omf_putdword(r,u);
	return 0;
}

int
omf_putname(orec_t *r,char *name, int len)
{
	omf_putbyte(r,len);
	while (len--)
		omf_putbyte(r,*name++);
	return 0;
}

int
omf_putindex(orec_t *r, unsigned index)
{
	if (index >= 0x80) {
		omf_putbyte(r,0x80 | (index>>8));
	}
	omf_putbyte(r, index);
	return 0;
}


int
omf_putc(int c, FILE *fp, int *csum)
{
	if (csum) {
		*csum += c & 0xff;
	}
	putc(c,fp);
	return c;
}

int
omf_putw(int w, FILE *fp, int *csum)
{
	omf_putc(w & 0xff,fp,csum);
	omf_putc((w >> 8) & 0xff, fp, csum);
	return 0;
}

int
omf_putl(long w, FILE *fp, int *csum)
{
	omf_putw(w & 0xffff, fp, csum);
	omf_putw((w>>16) & 0xffff, fp, csum);
	return 0;
}

int
omf_putrec(FILE *fp, orec_t *rec,int type)
{
	int	i;
	int	csum = 0;
	omf_putc(type, fp, &csum);
	omf_putw(rec->curp+1, fp, &csum);
	for (i=0; i < rec->curp; i++) {
		omf_putc(rec->bufp[i],fp, &csum);
	}
	csum &= 0xff;
	csum = 0x100 - csum;
#if defined DEBUG
{
	extern int verbose;
	if (verbose) {
		fprintf(stderr,"omf_putrec: {%x, %d %x}\n",
			type, rec->curp+2, csum);
	}
}
#endif
	omf_putc(csum,fp,0);
	return 0;
}
