/*- $Id: shobj.c,v 1.3 1998/03/19 19:15:50 steve Exp $
 * DESCRIPTION
 *  This module parses a specification file into a set of structures
 *  defining the shared library.
 */

#include <libc.h>
#include "ShObjDef.h"
#include "mkshlib.h"

#ifndef _argused
/* make the compiler think it is used */
#define _argused(x) ((x)=(x))
#endif

static const char WS[] = " \t\n";


typedef
struct ParserState {
	char         *fn;                 /* name of specification file */
	FILE         *fp;
	int           lineno;
	int           state;
	int           nerrs;
	int           nwarn;
	ShObjDef     *sh;
} ParserState;

static int default_init(ParserState *o, int argc, char **argv);
static int warning_body(ParserState *o, int argc, char **argv);
static int ignore_body(ParserState *o, int argc, char **argv);

static int default_done(ParserState *o);

static int nofloat_flag(ParserState *o, int argc, char **argv);
static int noinit_flag(ParserState *o, int argc, char **argv);
static int preserve_flag(ParserState *o, int argc, char **argv);
static int quiet_flag(ParserState *o, int argc, char **argv);
static int verbose_flag(ParserState *o, int argc, char **argv);

static int address_init(ParserState *o, int argc, char **argv);

static int size_init(ParserState *o, int argc, char **argv);

static int target_init(ParserState *o, int argc, char **argv);

static int ident_init(ParserState *o, int argc, char **argv);

static int externals_init(ParserState *o, int argc, char **argv);

static int objects_init(ParserState *o, int argc, char **argv);
static int object_body(ParserState *o, int argc, char **argv);
static int objnload_body(ParserState *o, int argc, char **argv);

static int init_init(ParserState *o, int argc, char **argv);
static int init_body(ParserState *o, int argc, char **argv);
//static int init_done(ParserState *o);

static int hide_init(ParserState *o, int argc, char **argv);
static int hide_body(ParserState *o, int argc, char **argv);
//static int hide_done(ParserState *o);

static int export_init(ParserState *o, int argc, char **argv);
static int export_body(ParserState *o, int argc, char **argv);
//static int export_done(ParserState *o);

static int branch_init(ParserState *o, int argc, char **argv);
static int branch_body(ParserState *o, int argc, char **argv);
//static int branch_done(ParserState *o);


/*-
 * In the initial version of this, we needed to use the "done"
 * routine to clean up lists; etc.
 * These lists have been changed to be one generic class which is
 * self maintained; so this is no longer necessary.
 * They are left in for later.
 * The enums much match the indices in Ptab[], so only add them when
 * necessary.
 */


enum {
	None,
	Error,
	Obj,
	ObjectsNoload,
	Init,
	Hide,
	Branch,
};

static
struct ParseEnt {
	char           *token;
	int           (*init)(ParserState *, int, char **);
	int           (*body)(ParserState *, int, char **);
	int           (*done)(ParserState *);
} Ptab[] = {

	{  "",           default_init,     warning_body,    default_done  },
	{  "",           default_init,     warning_body,    default_done  },
	{  "objects",    objects_init,     object_body,     default_done  },
	{  "objects",    objects_init,     objnload_body,   default_done  },
	{  "init",       init_init,        init_body,       default_done  },
	{  "hide",       hide_init,        hide_body,       default_done  },
	{  "branch",     branch_init,      branch_body,     default_done  },

	{  "address",    address_init,     warning_body,    default_done  },
	{  "export",     export_init,      export_body,     default_done  },
	{  "externals",  externals_init,   ignore_body,     default_done  },
	{  "hide",       hide_init,        hide_body,       default_done  },
	{  "ident",      ident_init,       warning_body,    default_done  },
	{  "size",       size_init,        warning_body,    default_done  },
	{  "target",     target_init,      warning_body,    default_done  },
	{  "nofloat",    nofloat_flag,     ignore_body,     default_done  },
	{  "noinit",     noinit_flag,      ignore_body,     default_done  },
	{  "preserve",   preserve_flag,    ignore_body,     default_done  },
	{  "quiet",      quiet_flag,       ignore_body,     default_done  },
	{  "verbose",    verbose_flag,     ignore_body,     default_done  },
	{  "",           default_init,     warning_body,    default_done  }
};

#define NSTATES (sizeof Ptab/sizeof Ptab[0])

#define SectInit(o, c, v)  (*(Ptab[(o)->state].init))(o, c, v)
#define SectBody(o, c, v)  (*(Ptab[(o)->state].body))(o, c, v)
#define SectDone(o)        (*(Ptab[(o)->state].done))(o)




static void
errmsg(ParserState *o, char *mstr, va_list varg)
{
	char	*bufp;
	fprintf(stderr,"%s(%d): %s", o->fn, o->lineno, mstr);
	while (bufp=va_arg(varg, char *))
		fputs(bufp, stderr);
	putc('\n', stderr);
	va_end(varg);
}

static void
parse_error(ParserState *o, ...)
{
	va_list     vargs;
	if (o->state == Error)
		return;
	va_start(vargs, o);
	errmsg(o, "error ", vargs);
	o->nerrs++;
	o->state = Error;
}

static void
parse_warning(ParserState *o, ...)
{
	va_list     vargs;
	if (o->state == Error)
		return;
	va_start(vargs, o);
	errmsg(o, "warning ", vargs);
	o->nwarn++;
}


static
iscomment(char *arg)
{
	return arg[0] == '#' && arg[1] == '#';
}


static
istypespec(char *arg)
{
	if (!iscomment(arg))
		return 0;
	arg +=2;
	if ((toupper(arg[0]) == 'F' && toupper(arg[1]) == 'U') ||
	    (toupper(arg[0]) == 'C' && toupper(arg[1]) == 'O') ||
	    (toupper(arg[0]) == '(' && toupper(arg[1]) == ')'))
		return 1;
	if ((toupper(arg[0]) == 'D' && toupper(arg[1]) == 'A') ||
	    (toupper(arg[0]) == '[' && toupper(arg[1]) == ']'))
		return 2;
	return 0;
}

static int
count_args(int argc, char **argv)
{
	int	i;
	for (i=0; i < argc; i++) {
		if (iscomment(argv[i]))
			break;
	}
	return i;
}

expect_args(ParserState *o, int argc, char **argv, int nexpect)
{
	int             n = count_args(argc, argv);
	if (n < nexpect) {
		parse_error(o, "missing argument near '", argv[n-1], "'", 0);
		return -1;
	}
	if (n > nexpect) {
		parse_error(o, "extra arguments near '", argv[nexpect+1], "'", 0);
		return -1;
	}
	return 0;
}
static int
is_cident(char *str)
{
	if (isalpha(*str) || *str == '_') {
		while (*str && (isalnum(*str) || *str == '_'))
			str++;
	}
	return !*str;
}

static int
get_symbol(char *type, char *srcn, char **ss, Symbol **sp)
{
	int          stype = 0;
	char        *s;
	char        *t;
	Symbol      *p;

	s = xstrdup(srcn, "get_symbol");
	if (t=strstr(s,"[]")) {
		stype = 2; 
		*t = '\0';
	} else if (t=strstr(s, "()")) {
		stype = 1;
		*t = '\0';
	} else if (type) {
		stype = istypespec(type);
	}
	if ((p=sym_lookup(s)) == 0) {
		p = new_symbol();
		if (stype == 2)
			p->is_data = 1;
		else if (stype == 1)
			p->is_code = 1;
		sym_install(s, p);
	} else if (!p->is_code && !p->is_data && stype) {
		p->is_code = stype == 1;
		p->is_data = stype == 2;
	}
	*sp = p;
	*ss = s;
	return 0;
}
/*-
 * command line switches may be set here
 */

static int
nofloat_flag(ParserState *o, int argc, char **argv)
{
	_argused(o);
	_argused(argc);
	_argused(argv);
	program_flags.nofloat = 1;
	return 0;
}
static int
noinit_flag(ParserState *o, int argc, char **argv)
{
	_argused(o);
	_argused(argc);
	_argused(argv);
	program_flags.set_init = 0;
	return 0;
}
static int
preserve_flag(ParserState *o, int argc, char **argv)
{
	_argused(o);
	_argused(argc);
	_argused(argv);
	program_flags.no_unlink = 1;
	return 0;
}
static int
quiet_flag(ParserState *o, int argc, char **argv)
{
	_argused(o);
	_argused(argc);
	_argused(argv);
	program_flags.quiet = 1;
	return 0;
}
static int
verbose_flag(ParserState *o, int argc, char **argv)
{
	_argused(o);
	_argused(argc);
	_argused(argv);
	program_flags.print_stats = 1;
	diag_level(DIAG_DEBUGLEVEL);
	return 0;
}
/*-
 * #address [ .text | .data ] address
 * There is no body for this section
 */
static int
address_init(ParserState *o, int argc, char **argv)
{
	long          addr;

	if (expect_args(o, argc, argv, 3) == -1)
		return 0;

	addr = strtol(argv[2],0,0);	
	if (strcmp(argv[1],".text") == 0) {
		o->sh->text_addr = addr;
	} else if (strcmp(argv[1],".data") == 0) {
		o->sh->data_addr = addr;
	} else {
		parse_error(o, argv[2], " expected .text or .data", 0);
	}
	return 0;
}
/*-
 * #size [.data | .text]? size
 * There is no body for this section
 */

static int
size_init(ParserState *o, int argc, char **argv)
{
	if (expect_args(o, argc, argv, 2) == -1)
		return 0;
	o->sh->image_size = strtol(argv[1],0,0);
	return 0;
}

static int
default_init(ParserState *o, int argc, char **argv)
{
	_argused(argc);
	parse_error(o, "Unknown section near ", argv[0], 0);
	return 0;
}
/*-
 * default action for a body within a section which doesn't expect it
 * is to raise an error
 */
static int
warning_body(ParserState *o, int argc, char **argv)
{
	_argused(argc);
	parse_warning(o, "expected section begin near ", argv[0], 0);
	return 0;
}
static int
ignore_body(ParserState *o, int argc, char **argv)
{
	_argused(o);
	_argused(argc);
	_argused(argv);
	return 0;
}


static int
default_done(ParserState *o)
{
	_argused(o);
	return 0;
}
/*-
 * #target target-name
 * this section has no body
 */

static int
target_init(ParserState *o, int argc, char **argv)
{
	if (expect_args(o, argc, argv, 2) == -1)
		return 0;
	if (o->sh->target == 0)
		o->sh->target = xstrdup(argv[1], "parse::target_init");
	else
		parse_warning(o, "redefinition of target ignored\n", 0);
	return 0;
}
/*-
 * #ident name
 * This section has no body and is ignored
 */
static int
ident_init(ParserState *o, int argc, char **argv)
{
	if (expect_args(o, argc, argv, 2) == -1)
		return 0;
	return 0;
}
/*-
 * externals *
 * body is ignored
 */
static int
externals_init(ParserState *o, int argc, char **argv)
{
	_argused(argv);
	_argused(argc);
	parse_warning(o, "ignoring section #externals\n", 0);
	return 0;
}

/*-
 * #objects [noload]?
 * This section is either a set of loadable or non-loadable object
 * modules.
 */

static int
objects_init(ParserState *o, int argc, char **argv)
{
	int        nargs = count_args(argc, argv);

	if (nargs == 2) {
		if (strcmp(argv[1], "noload") == 0) {
			o->state = ObjectsNoload;
		} else {
			parse_error(o, argv[1], " expected \"noload\"", 0);
		}
	} else {


	}
	return 0;
}


static int
object_body(ParserState *o, int argc, char **argv)
{
	if (expect_args(o, argc, argv, 1) == -1)
		return 0;
	put_nlist(&o->sh->objs, xstrdup(argv[0], "object_body"));
	return 0;
}

static int
objnload_body(ParserState *o, int argc, char **argv)
{
	if (expect_args(o, argc, argv, 1) == -1)
		return 0;
	/* put name onto object-noload file list here */
	put_nlist(&o->sh->obj_nol, xstrdup(argv[0], "object_body"));
	return 0;
}


/*-
 * #init object-file
 * This section consists of a list of name pairs.
 */

static int
init_init(ParserState *o, int argc, char **argv)
{
	if (expect_args(o, argc, argv, 2) == -1)
		return 0;
	/* here store init file name */
	return Init;
}

static int
init_body(ParserState *o, int argc, char **argv)
{
	int          nargs = count_args(argc, argv);
	char        *symn;
	Symbol      *p;

	if (nargs < 2) {
		return 0;
	}

	if (!is_cident(argv[0])) {
		parse_error(o, argv[0], " Is not a valid C identifier!\n", 0);
		return -1;
	}
	get_symbol(argv[2], argv[1], &symn, &p);
	if (!is_cident(symn)) {
		parse_error(o, symn, " Is not a valid C identifier!\n", 0);
		return -1;
	}
	put_nlist(&o->sh->init[0], xstrdup(argv[0],"init_body"));
	put_nlist(&o->sh->init[1], symn);
	return 0;
}

static int
hide_init(ParserState *o, int argc, char **argv)
{
	int         nargs;
	nargs = count_args(argc, argv);
	if (argc < 2) {
		parse_error(o, argv[0], " expected symbol name or pattern", 0);
		return 0;
	}
	if (strcmp(argv[1], "linker")) {
		parse_error(o, argv[1], " expected specifier \"linker\"", 0);
		return 0;
	}
	if (argc > 2) {
	/* store each string into hide list */
		put_nlist(&o->sh->hide, xstrdup(argv[2], "hide_init"));
	}
	return Hide;
}
static int
hide_body(ParserState *o, int argc, char **argv)
{
	int         nargs;
	nargs = count_args(argc, argv);
	if (argc > 1) {
	}
	put_nlist(&o->sh->hide, xstrdup(argv[0],"hide_body"));
	return 0;
}

static int
export_init(ParserState *o, int argc, char **argv)
{
	if (expect_args(o, argc, argv, 2) == -1)
		return 0;

	if (strcmp(argv[1], "linker"))  {
		parse_error(o, argv[1]," expected string \"linker\"", 0);
		return 0;
	}
	/* store object file name here */
	return 0;
}



static int
export_body(ParserState *o, int argc, char **argv)
{
	char       *symn;
	Symbol     *s;
	_argused(argc);
	get_symbol(argv[1], argv[0], &symn, &s);
	if (!is_cident(symn)) {
		parse_error(o, symn, " is not a valid C identifier", 0);
		return -1;
	}
	s->is_exported = 1;

	put_nlist(&o->sh->export, symn);
	return 0;
}


static int
branch_init(ParserState *o, int argc, char **argv)
{
	_argused(o);
	_argused(argc);
	_argused(argv);
	return Branch;
}

static
branch_add(ParserState *o, char *nm, long r0, long r1)
{
	ShObjDef     *sh = o->sh;
	NameList     *br = &sh->branch;
	Symbol       *p;

	if (r1 <= r0) {
		parse_error(o, "Invalid branch range near ", nm, 0);
		return -1;
	}
	if (r0 > br->count) {
		r0 = br->count;
	} else if (r0 < br->count) {
		parse_error(o, "Conflicting branch indices near ", nm, 0);
		return -1;
	}
	get_symbol(0, nm, &nm, &p);
	if (IS_DATA(p)) {
		parse_error(o, "Conflicting definitions for symbol ", nm, 0);
		p->is_data = 0;
		p->is_sect = 0;
		p->is_exported = 0;
	}
	if (!IS_BRANCH(p)) {
		p->branch_index = r0+1;
		p->is_branch = 1;
		p->is_code   = 1;
	}
	while (r0 < r1) {
		put_nlist(br, nm);
		r0++;
	}
	return 0;
}


int
branch_body(ParserState *o, int argc, char **argv)
{
	long       r0,
	           r1;
	char      *p;
	int        i;
	int        nargs;

	if ((nargs=count_args(argc, argv)) < 2) {
		parse_error(o, argv[0], "malformed branch specification",0);
		return 0;
	}
	r0 = strtol(argv[1], &p, 0);
	if (p == argv[i=1]) {
		parse_error(o, argv[1], "branch index must be numeric",0);
		return 0;
	}
	if (!*p) {	/* end of string */
		i++;
		if (argc < 3 || iscomment(argv[i])) {
			branch_add(o, argv[0], r0-1, r0);
			return 0;
		}
		p = argv[i];
	}
	if (*p != '-') {
		parse_error(o, argv[i]," expected '-' for branch range", 0);
		return 0;
	}
	if (!*++p) {
		i++;
		if (argc < i+1 || iscomment(argv[i])) {
			parse_error(o, argv[i-1]," expected '-' for branch range", 0);
			return 0;
		}
		p = argv[i];
	}
	r1 = strtol(p, &p, 0);
	branch_add(o, argv[0], r0-1, r1-1);
	return 0;
}

int
ShObjDef_Init(ShObjDef *o)
{
	memset(o, 0, sizeof *o);
	init_nlist(&o->branch);
	init_nlist(&o->objs);
	init_nlist(&o->obj_nol);
	init_nlist(&o->init[0]);
	init_nlist(&o->init[1]);
	init_nlist(&o->hide);
	init_nlist(&o->export);

	o->text_addr = 0;
	o->text_len = 0;
	o->data_addr = 0;
	o->data_len = 0;
	o->image_size = 0;
	o->jmptab_offs = 0x1000;
	o->target    = 0;
	o->host      = 0;
	return 0;
}

int
ShObjDef_Load(ShObjDef *shp, char *fn, FILE *fp)
{
	char            buf[LINE_MAX];
#define AMAX 50
	char           *argv[AMAX];
	int             argc;

	ParserState     p;
	p.nwarn  = 0;
	p.nerrs  = 0;
	p.lineno = 0;
	p.fn     = fn;
	p.fp     = fp;
	p.sh     = shp;
	p.state  = None;

	while (fgets(buf, sizeof buf, p.fp)) {
		p.lineno++;
		if ((argc = breakstr(buf, WS, argv, AMAX)) == 0)
			continue;

		if (argv[0][0] == '#' && argv[0][1] == '#')
			continue;
		if (argv[0][0] == '#') {
			int           i;
			SectDone(&p);
			for (i=0; i < NSTATES &&
				strcmp(Ptab[i].token, argv[0]+1);
				i++);
			if (i == NSTATES) {
				p.state = Error;
			} else {
				p.state = i;
			}
			SectInit(&p, argc, argv);
		} else {
			SectBody(&p, argc, argv);
		}
	}
	SectDone(&p);
	return p.nerrs ? -1 : 0;
}
