/*-
Module:      target
Purpose:     Build target shared library.
Inputs:      Shared Library Specification;
Outputs:
         1)   Target Shared Library.
         2)   Addresses of Symbols in Target Library.

Processing:
      - Generate an executable from target object files.
      - Due to wlink problems; this is done twice -- the first time
      is used to get the size of the .text segment
      An object module is added which pads the .text out to where the
      .data segment can be placed.

*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "mkshlib.h"
#include "codegen.h"
#include "ldfile.h"

/* for use in error messages */
#define MODULE_NAME	"generate target"


BldMain(FILE * f, ShObjDef *sh)
{
	char           *src;
	char           *dest;
	FILE           *fp;
	char           *mmac;

	if ((mmac = getmac("MainText")) == 0) {
		fatalerr("Missing macro for MainText");
		return -1;
	}
	if ((src = temp(TEMP_TARGET, ".c")) == 0) {
		fatalerr("BldMain, cannot get temp file");
	}
	if ((dest = temp(TEMP_TARGET, ".o")) == 0) {
		untemp(src,program_flags.no_unlink);
		fatalerr("BldMain, cannot get temp file");
	}
	diag_msg(DIAG_DEBUGLEVEL, "%s is the source for 'main()' in the target\n", src);
	diag_msg(DIAG_DEBUGLEVEL, "%s is the object for %s\n", dest, src);

	if ((fp = xfopen(src, "w", "BldMain")) == 0) {
		untemp(src, program_flags.no_unlink);
		untemp(dest, program_flags.no_unlink);
		return -1;
	}
	fmacrov(fp, mmac, sh->prefix, 0);
	fclose(fp);

	cc(src, dest, "");
	untemp(src, program_flags.no_unlink);

	if ((fp = xfopen(dest, "r", "BldMain")) == 0) {
		untemp(dest,program_flags.no_unlink);
		return -1;
	} else {
		fcat(f, fp);
		fclose(fp);
		untemp(dest,program_flags.no_unlink);
		return 0;
	}
}


/*-
 * fix_dseg -- look at the target and generate an object file which
 * fixes the base address of the data segment
 */

long
AdjustDataSeg(ShObjDef *sh, char *src, char *target)
{
	LdInfo          ldin;
	long            lsize;
	ulong_t         data_addr = sh->data_addr;

	if (ldf_seglen(target, &ldin) == -1) {
		return -1;
	}
	if (program_flags.print_stats) {
		fprintf(stderr, ".text is 0x%lx bytes\n", ldin.ld_tsize);
		fprintf(stderr, ".data is 0x%lx bytes\n", ldin.ld_dsize);
		fprintf(stderr, ".bss  is 0x%lx bytes\n", ldin.ld_bsize);
	}
	if (data_addr) {
		FILE           *fp;
		long            padlen = data_addr -
		                        (ldin.ld_vaddr + ldin.ld_tsize);
		if (padlen < 0) {
			warning(".text too large; .data begins at $%x, "
				".text ends at $%x\n",
				data_addr, ldin.ld_vaddr + ROUND_PAGE(ldin.ld_tsize));
			return -1;
		}
		if ((fp = xfopen(src, "a+", "AdjustDataSeg")) == 0) {
			fatalerr("cannot open object for write");
		}
		mktextpad(fp, padlen);
		fclose(fp);
		ldin.ld_tsize += padlen;
	} else {
		warning(".data section not defined, assuming default");
	}
	lsize = ROUND_PAGE(ldin.ld_tsize) +
		ROUND_PAGE(ldin.ld_dsize + ldin.ld_bsize);

	if (lsize > sh->image_size) {
		if (sh->image_size) {
			warning("target requires %ld bytes - %ld specified --"
                                "new target may not be compatible\n",
                                	lsize, sh->image_size);
                }
		sh->image_size = lsize;
	}
	return 0;
}

static
BldJumpTable(FILE *f, ShObjDef *sh)
{
	struct jmptab   jp;
	int             i;
	NameList        *nm = &sh->branch;

	jp.jl_modnam = "JumpTable";
	jp.j_nvec = nm->count;
	jp.j_vec = xcalloc(sizeof *jp.j_vec, jp.j_nvec + 1, "JumpTable");

	for (i=0; i < nm->count; i++) {
		jp.j_vec[i] = nm->names[i];
	}
	jp.j_vec[i] = 0;
	jp.jl_start  = sh->tar_names.exp0;
	jp.jl_end    = sh->tar_names.exp1;
	jp.jl_relocp = sh->tar_names.bind;
	jp.j_offset  = sh->jmptab_offs;
	mkjumptable(f, &jp);
	return 0;
}

static
AddObjects(FILE * f, ShObjDef *sh)
{
	NameList          *nm;
	int                i;

	nm = &sh->objs;
	for (i=0; i < nm->count; i++) {
		FILE           *fp;
		if (fp = xfopen(nm->names[i], "r", "AddObjects")) {
			fcat(f, fp);
			fclose(fp);
		}
	}
	return 0;
}

static
BldInitTable(FILE * f, ShObjDef *sh)
{

	char           *src, *dest;
	FILE           *fp;
	char            tempstr[32];
	char           *tarmac = getmac("TargetImport");
	NameList       *nm;
	int             i;


	if (tarmac == 0) {
		fatalerr("missing macro for Target Import Definition\n");
	}
	if ((src = temp(TEMP_TARGET, ".c")) == 0) {
		fatalerr("BldInitTable: cannot get temp file");
	}
	if ((dest = temp(TEMP_TARGET, ".o")) == 0) {
		untemp(src,program_flags.no_unlink);
		fatalerr("BldInitTable: cannot open temp file");
	}
	diag_msg(DIAG_DEBUGLEVEL, "%s is the source for the init list in the target\n", src);
	diag_msg(DIAG_DEBUGLEVEL, "%s is the object for %s\n", dest, src);

	if ((fp = xfopen(src, "w", "BldInitTable")) == 0) {
		fatalerr("BldInitTable: cannot open src file");
	}
	nm = &sh->init[0];
	for (i=0;  i < nm->count; i++) {
		fprintf(fp, "int %s = -1;\n", nm->names[i]);
	}
	fmacrov(fp, tarmac,
		sh->tar_names.bind,
		sh->prefix,
		sh->tar_names.exp0,
		nm->count ? nm->names[0] : "_nullarea",
		ltoa(nm->count, tempstr, 10),
		0);
	fclose(fp);
	cc(src, dest, "");
	untemp(src,program_flags.no_unlink);
	if ((fp = xfopen(dest, "r", "BldInitTable")) == 0) {
		untemp(dest,program_flags.no_unlink);
		return -1;
	}
	fcat(f, fp);
	fclose(fp);
	untemp(dest,program_flags.no_unlink);

	for (i=0; i < sh->init[1].count; i++) {
		Symbol *p = sym_lookup(sh->init[1].names[i]);
		if (p && !p->is_data && ! IS_BRANCH(p))
			mk_jump(f, sh->init[0].names[i], sh->init[1].names[i]);
	}
	return 0;
}

BldTarget(ShObjDef *sh)
{
	char           *s;
	FILE           *f;

	if (!(s = temp(TEMP_TARGET, ".o"))) {
		fatalerr("mktarget: cannot get temp file");
	}
	diag_msg(DIAG_DEBUGLEVEL, "%s is the container for the target objects\n", s);

	if ((f = xfopen(s, "w", "mktarget")) == 0) {
		fatalerr("make target unable to create %s:%s\n", s, strerror(errno));
	}

	BldInitTable(f, sh);
	BldMain(f, sh);
	AddObjects(f, sh);
	BldJumpTable(f, sh);
	mketext(f);
	fclose(f);

	ld(s, sh->target, loader_opts, sh->text_addr);

	AdjustDataSeg(sh, s, sh->target);

	ld(s, sh->target, loader_opts, sh->text_addr);
	untemp(s,program_flags.no_unlink);
	return 0;
}


static char dump_hdr_str[] = "%-10.10s\t%-10.10s\t%-10.10s\t%-10.10s\t%s\n";
static char dump_fmt_str[] = "%-10.10s\t% 10.10ld\t%-10.10s\t0x%-08.8lx\t%s\n";

static
print_symbol(char *name, Symbol *p, void *q)
{
	_notused(q);
	if (!p) {
		diag_msg(DIAG_INFOLEVEL, dump_fmt_str, "unknown",
			0, "unknown",
			0, name);
		diag_msg(DIAG_INFOLEVEL, "anon\t0\tprivate\t0\t*%s\n",
			name);
	} else {
		diag_msg(DIAG_INFOLEVEL, dump_fmt_str,
		 	p->is_code ? "code" : p->is_data ? "data" : "anon",
		 	p->branch_index,
		 	p->is_exported ? "exported" : "private",
		 	p->address, name);
	}
}

static
print_symtab()
{
	diag_msg(DIAG_INFOLEVEL, "Symbol Table:\n");
	diag_msg(DIAG_INFOLEVEL, dump_hdr_str, "Type", "Branch", "Linkage", "Address", "Name");
	sym_apply(print_symbol, 0);
	return 0;
}

static
check_symbol(char *name, Symbol * p, int *nundef)
{

	if (IS_EXPORTED(p) && p->address == 0) {
		error("exprted%s symbol \"%s\" is not defined",
		      IS_BRANCH(p) ? " (branch)" : "", name);
		++*nundef;
	}
	return 0;
}

int
sym_setaddr(LdfName * nl, int *nerrs)
{
	Symbol         *sym;
	char            namebuf[1024];

	/*-
	 * adjust the name to a "C" symbol.
	 */
	if (nl->type & _SY_DATA) {
		strcpy(namebuf, nl->name + 1);
	} else {
		strcpy(namebuf, nl->name);
		namebuf[strlen(namebuf) - 1] = '\0';
	}

	if (sym = sym_lookup(namebuf)) {
		if (IS_CODE(sym) && (nl->type & _SY_DATA)) {
			error("\"%s\" is defined as code, was found in data segment\n",
			      namebuf);
			++*nerrs;
		} else if (IS_DATA(sym) && (nl->type & _SY_CODE)) {
			error("\"%s\" is defined as data, was found in code segment\n",
			      namebuf);
			++*nerrs;
		}
		sym->address = nl->offset;
	}
	return 0;
}

/*-
 * mktarget.
 * 1. Build the target library.
 * 2. Augment symbol table from symbol information in target.
 * 3. Check for type errors between target and symbol table.
 * 4. Check for undefined symbols in target.
 *
 */
mktarget(ShObjDef *sh)
{
	LdfName        *namelist;
	int             errcount;

	BldTarget(sh);

	if ((namelist = ldf_nlist(sh->target)) == 0) {
		unlink(sh->target);
		fatalerr("could not extract name list target");
	}
	errcount = 0;
	ldf_walk_nlist(namelist, sym_setaddr, &errcount);
	ldf_del_nlist(namelist);
	if (errcount) {
		unlink(sh->target);
		fatalerr("%d type errors detected", errcount);
	}

	print_symtab();

	errcount = 0;
	sym_apply(check_symbol, &errcount);
	if (errcount) {
		unlink(sh->target);
		fatalerr("%d undefined exported symbols were discovered", errcount);
	}
	return 0;
}
