
/*-
miscellaneous routines which "cover" base ones.

The open/alloc ones print a canned message with a provided string which should
say something about the origin.

temp()
creates a temporary name with a given suffix.

fcat()
copies from one file into another.

*/

#include <libc.h>
#include "mkshlib.h"

extern char *program_name;
#define PROGNAME (program_name ? program_name : "PROGRAM??: ")

static
err_prefix(void)
{
	if (program_name) {
		fputs(program_name, stderr);
		fputs(": ", stderr);
	}
}



void 
fatalerr(char *fmt,...)
{
	va_list         varg;
	va_start(varg, fmt);
	err_prefix();
	fputs("fatal error ", stderr);
	vfprintf(stderr, fmt, varg);
	putc('\n', stderr);
	exit(1);
}
void
warning(char *fmt,...)
{
	va_list         varg;
	va_start(varg, fmt);
	fputs("mkshlib warning: ", stderr);
	vfprintf(stderr, fmt, varg);
	putc('\n', stderr);
}
void
error(char *fmt,...)
{
	va_list         varg;
	va_start(varg, fmt);
	fputs("mkshlib error: ", stderr);
	vfprintf(stderr, fmt, varg);
	putc('\n', stderr);
}

int
fcat(FILE * dest, FILE * src)
{
	int             c;
	int             n = 0;
	for (n = 0; (c = getc(src)) != EOF; n++)
		putc(c, dest);
	return n;
}

#undef unlink
int
xunlink(char *name, const char *func)
{
	if (unlink(name) == -1) {
		int       y = errno;
		err_prefix();
		fprintf(stderr, "could not remove file '%s' (%s) in %s\n",
			name, strerror(y), func);
		return -1;
	}
	return 0;
}

#undef fopen
FILE           *
xfopen(char *nam, const char *mode, const char *func)
{
	FILE           *f;
	if ((f = fopen(nam, mode)) == NULL) {
		err_prefix();
		fprintf(stderr, "error opening file '%s' in '%s': %s\n",
			nam, func, strerror(errno));
	}
	return f;
}
#undef calloc
void           *
xcalloc(int nelem, int elsize, char *func)
{
	void           *p;
	if (!(p = calloc(nelem, elsize))) {
		err_prefix();
		fprintf(stderr, "error allocating %d %d's in '%s': %s\n",
			nelem, elsize, func, strerror(errno));
	}
	return p;
}

void           *
xmalloc(int nelem, char *func)
{
	return xcalloc(1, nelem, func);
}

#undef realloc
void           *
xrealloc(void *p, int msize, char *func)
{
	void           *x;
	if ((x = realloc(p, msize)) == 0) {
		err_prefix();
		fprintf(stderr, "error reallocating %x to size %d in '%s': %s\n",
			p, msize, func, strerror(errno));
	}
	return x;
}
#undef strdup
char *
xstrdup(char *src, const char *func)
{
	char           *p;
	if ((p=strdup(src)) == 0) {
		fatalerr("duplicating string '%s' in '%s' : %s\n",
			src, func, strerror(errno));
	}
	return p;
}

#define MAX_TFILE 256
static int  tcounter = 0;
static char  *tmplist[MAX_TFILE];

static void
temp_cleanup(void)
{
	int             i;
	for (i=0; i < tcounter; i++) {
		if (tmplist[i]) {
			xunlink(tmplist[i], "temp_cleanup");
			free(tmplist[i]);
			tmplist[i] = 0;
		}
	}
}

char *
temp(char *prefix, const char *postfix)
{
	char            buf[PATH_MAX];
	int             fd;
	char           *p;
	char           *tdir;

	if (tcounter == 0) {
		atexit(temp_cleanup);
	}

	if ((tdir=getenv("TMPDIR")) == 0) {
		tdir = "/tmp";
	}

	for (p=buf; *p++ = *tdir++; );
	p[-1] = '/';
	for (; *p++ = *prefix++;);
	ltoa(getnid(), p-1, 10);
	p += strlen(p);
	*p++ = '-';
	ltoa(getpid(), p, 10);
	p += strlen(p);
	*p++ = '-';

	do {
		char       *t = p;
		ltoa(tcounter, t, 10);
		t += strlen(t);
		for (; *t++ = *postfix++;);
		tcounter++;
	} while ((fd=creat(buf, 0644)) == -1 && errno == EEXIST);

	close(fd);
	if ((p=xmalloc(strlen(buf)+1, "temp")) == 0) {
		unlink(buf);
		return 0;
	}
	strcpy(p, buf);
	return tmplist[tcounter-1] = p;
}

int
untemp(char *name, int no_unlink)
{
	int             i;
	for (i=0; i < tcounter && tmplist[i] != name; i++)
		;
	if (i == tcounter)
		return -1;
	if (no_unlink == 0)
		xunlink(name, "untemp");
	free(name);
	tmplist[i] = 0;
	return 0;
}
